/* Copyright (c) 2023 Skyward Experimental Rocketry
 * Author: Nicolò Caruso
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#pragma once

#include <diagnostic/PrintLogger.h>
#include <drivers/spi/SPIDriver.h>
#include <drivers/timer/TimestampTimer.h>
#include <miosix.h>
#include <sensors/Sensor.h>

#include "MY_LIS3DSHData.h"
#include "MY_LIS3DSHDefs.h"

namespace Boardcore
{
/*! MY_LIS3DSH class for the LIS3DSH sensor */
class MY_LIS3DSH : public Sensor<MY_LIS3DSHData>
{
public:
    /**
     * @brief Constructor without spi configuration
     *
     * @param spiBus : SPI bus of the discovery, connected to the sensor
     * @param cs : Sensor's Chip Select
     * @param odr : Sensor's Output Data Rate
     * @param bdu : Sensor's Block Data Update
     * @param fs : Sensor's Full Scale Range
     * */
    MY_LIS3DSH(SPIBusInterface& spiBus, miosix::GpioPin cs,
               MY_LIS3DSHDefs::OutputDataRate odr,
               MY_LIS3DSHDefs::BlockDataUpdate bdu,
               MY_LIS3DSHDefs::FullScaleRange fs  // Full scale range
    );

    /**
     * @brief Constructor with spi configuration
     *
     * @param spiBus : SPI bus of the discovery, connected to the sensor
     * @param cs : Sensor's Chip Select
     * @param cfg : SPI bus configuration
     * @param odr : Sensor's Output Data Rate
     * @param bdu : Sensor's Block Data Update
     * @param fs : Sensor's Full Scale Range
     * */
    MY_LIS3DSH(SPIBusInterface& spiBus, miosix::GpioPin cs, SPIBusConfig cfg,
               MY_LIS3DSHDefs::OutputDataRate odr,
               MY_LIS3DSHDefs::BlockDataUpdate bdu,
               MY_LIS3DSHDefs::FullScaleRange fs /*! Full scale range */
    );

    static SPIBusConfig
    getDefaultSPIConfig(); /*! Gets the default SPI configuration */

    /**
     * @brief Initialization function for the LIS3DSH sensor
     * It does configure the registers
     * @returns True if the initialization has been completed correctly
     * @returns False otherwise
     */
    bool init();

    /**
     * @brief Samples the data from the sensor's registers.
     *
     * Converts the data read from spi to an MY_LIS3DSHData structure
     *
     * @returns A copy of the last sampled data of MY_LIS3DSHData type.
     */
    MY_LIS3DSHData sampleImpl() override;

    /**
     * @brief Simply returns true.
     *
     * @returns True, always.
     */
    bool selfTest();

private:
    /*! SPI driver for transactions */
    SPISlave spi;
    // SPIDriver spi_driver;

    /*! odr, bdu, fs get from the constructor. */
    MY_LIS3DSHDefs::OutputDataRate odr;
    MY_LIS3DSHDefs::BlockDataUpdate bdu;
    MY_LIS3DSHDefs::FullScaleRange fs;
    bool initialized;
    PrintLogger logger = Logging::getLogger("li3dsh");
};

}  // namespace Boardcore
