/* Copyright (c) 2023 Skyward Experimental Rocketry
 * Author: Nicolò Caruso
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

namespace Boardcore
{
namespace MY_LIS3DSHDefs
{
// Given by the datasheet FS ranges
// Bitwise operation, to have 00, 01. 10, 11 representation
// Offset given by its position into the CTRL_REG4
enum FullScaleRange
{
    FS_02 = 0 << 3,  // +- 2.0 sensitivity
    FS_04 = 1 << 3,  // +- 4.0 sensitivity
    FS_06 = 2 << 3,  // +- 6.0 sensitivity
    FS_08 = 3 << 3,  // +- 8.0 sensitivity
    FS_16 = 4 << 3   // +- 16.0 sensitivity
};

// Offset given by the register CTRL_REG1 table!

enum OutputDataRate
{
    ODR_PD    = 0b0000 << 4,  //< Power down
    ODR_3_125 = 0b0001 << 4,  //< 3.125 Hz ODR
    ODR_6_25  = 0b0010 << 4,  //<  6.25 Hz ODR
    ODR_12_5  = 0b0011 << 4,  //<  12.5 Hz ODR
    ODR_25    = 0b0100 << 4,  //<    25 Hz ODR
    ODR_50    = 0b0101 << 4,  //<    50 Hz ODR
    ODR_100   = 0b0110 << 4,  //<   100 Hz ODR
    ODR_400   = 0b0111 << 4,  //<   400 Hz ODR
    ODR_800   = 0b1000 << 4,  //<   800 Hz ODR
    ODR_1600  = 0b1001 << 4,  //<  1600 Hz ODR
};

// BDU mode for update, given by position in register CTRL_REG4
enum BlockDataUpdate
{
    BDU_CONTINOUS_UPDATE = 0 << 3,
    BDU_WAIT_UNTIL_READ  = 1 << 3
};

/*! Self test enumeration */
enum SelfTest
{
    ST_NORMAL   = 0 << 1,
    ST_POSITIVE = 1 << 1,
    ST_NEGATIVE = 2 << 1
};

// Registers position, given by the register mapping of the datasheet
enum Registers
{
    REG_WHO_AM_I   = 0x0F,  //< Who Am I register
    REG_CTRL_REG1  = 0x21,
    REG_CTRL_REG2  = 0x22,
    REG_CTRL_REG3  = 0x23,
    REG_CTRL_REG4  = 0x20,
    REG_CTRL_REG5  = 0x24,
    REG_CTRL_REG6  = 0x25,
    REG_STATUS_REG = 0x27,
    REG_OUT_X_L    = 0x28,
    REG_OUT_X_H    = 0x29,
    REG_OUT_Y_L    = 0x2A,
    REG_OUT_Y_H    = 0x2B,
    REG_OUT_Z_L    = 0x2C,
    REG_OUT_Z_H    = 0x2D
};

enum Statuses : uint8_t
{
    /*  Register if Data Available */
    STATUS_REG_XDA   = 0b0000'0001,  // Data available on X-axis
    STATUS_REG_YDA   = 0b0000'0010,  // Data available on Y-axis
    STATUS_REG_ZDA   = 0b0000'0100,  // Data available on Z-axis
    STATUS_REG_XYZDA = 0b0000'1000,  // Data available on X-axis
    /* Registers for data overrnun (x overwrite) */
    STATUS_REG_XOR   = 0b0001 << 4,
    STATUS_REG_YOR   = 0b0010 << 4,
    STATUS_REG_ZOR   = 0b0100 << 4,
    STATUS_REG_XYZOR = 0b1000 << 4,
};

// Given by the CTRL_REG1 refernce register
constexpr uint8_t CTRL_REG4_XEN = 0b1;       // X axis Enabled
constexpr uint8_t CTRL_REG4_YEN = 0b1 << 1;  // Y axis Enabled
constexpr uint8_t CTRL_REG4_ZEN = 0b1 << 2;  // Z axis Enabled

/**
 * @brief make automatically increase the register addr. reading bytes with SPI
 */
constexpr uint8_t AUTOINC_ADDR = 0b0100'0000;

// Given by the datasheet register
constexpr uint8_t WHO_AM_I_ID = 0x3F;  // Here it is 00111111 -> 0x3F

// Sensitivity values from the mechanical characteristics (Table 4 pag.10)
// Here there are 3 different values for the 3 power modes Low power, Normal,
// High Rate
constexpr float SENSITIVITY_VALUES[] = {0.06, 0.12, 0.18, 0.24, 0.73};

}  // namespace MY_LIS3DSHDefs
}  // namespace Boardcore