/* Copyright (c) 2023 Skyward Experimental Rocketry
 * Author: Nicolò Caruso
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <drivers/spi/SPIDriver.h>
#include <drivers/timer/TimestampTimer.h>
#include <miosix.h>
#include <sensors/MY_LIS3DSH/MY_LIS3DSH.h>
#include <utils/Debug.h>

using namespace Boardcore;
using namespace miosix;

/*! SPI bus + pin configuration for the stm32f407vg discovery*/
SPIBus spiBus(SPI1);
/*! These parameters are given by the intersection of the Alternate Function and
 * Pins*/
GpioPin cs(GPIOE_BASE, 3);     /*< The chip select was PIN_E nr. 3*/
GpioPin spiSck(GPIOA_BASE, 5); /*< f_Sck at most 42 MHz*/
GpioPin spiMiso(GPIOA_BASE, 6);
GpioPin spiMosi(GPIOA_BASE, 7);

int main()
{
    Thread::sleep(1000);

    spiSck.mode(miosix::Mode::ALTERNATE);
    spiSck.alternateFunction(5); /*!< This... sets the alternate function to 5*/
    spiSck.speed(Speed::_100MHz);

    spiMiso.mode(miosix::Mode::ALTERNATE);
    spiMiso.alternateFunction(5);
    spiMiso.speed(Speed::_100MHz);

    spiMosi.mode(miosix::Mode::ALTERNATE);
    spiMosi.alternateFunction(5);
    spiMosi.speed(Speed::_100MHz);

    /*! Prepare for writing into the SPIbus */
    cs.mode(miosix::Mode::OUTPUT);
    cs.high();

    Thread::sleep(100);

    MY_LIS3DSH lis3dsh(
        spiBus, cs, MY_LIS3DSH::getDefaultSPIConfig(), MY_LIS3DSHDefs::ODR_100,
        MY_LIS3DSHDefs::BDU_WAIT_UNTIL_READ, MY_LIS3DSHDefs::FS_02);

    /*! If not init() correctly print error, return -1 */
    if (!lis3dsh.init())
    {
        lis3dsh.getLastError() == SensorErrors::INVALID_WHOAMI
            ? printf("ERROR - Failed init due to incorrect WHOAMI!\n")
            : printf("ERROR - Failed init!\n");
        return -1;
    }
    printf("Init successful\n");

    printf("Performing self test...\n");
    if (!lis3dsh.selfTest())
    {
        printf("ERROR - Failed self test\n");
    }

    printf("Self test successful\n");

    MY_LIS3DSHData sample;

    printf("Sampling...\n");
    /*! Sample data from the sensor for fixed iterations*/
    for (int i = 0; i < 100; i++)
    {
        lis3dsh.sample();
        sample = lis3dsh.getLastSample();
        TRACE("SAMPLE %d time %llu: %f, %f, %f\n", i,
              sample.accelerationTimestamp, sample.accelerationX,
              sample.accelerationY, sample.accelerationZ);
        Thread::sleep(100);
    }

    return 0;
}